Require Import Classical.
Require Import List.
Require Import Omega.
Require Import Program.
Require Import Paco.paco.
Import ListNotations.

Parameter State : Type.
Parameter trans : State -> State -> Prop.


(*symbolic states *) 
Definition SymState := State -> Prop.
Definition top : SymState := fun _ => True.
Definition bot : SymState := fun _ => False.
Definition cnj(p q :SymState) := fun s => p s /\ q s.
Definition dsj(p q :SymState) := fun s => p s \/ q s.
Definition neg (p:SymState) := fun s => ~ p s.
Definition imp(p q :SymState) := forall s, p s -> q s.

(* symbolic transition function *)
Definition symTrans(q:SymState) := fun s => exists s', q s' /\ trans s' s.

(* final states *)
Definition final : SymState := fun s => forall s', ~trans s s'.

(* sequences and paths *)
CoInductive seq  : Type :=
| one : forall s:State , seq
| add : State -> seq  -> seq.                                          
Definition head(sq : seq) : State :=
  match sq with
    | one s => s
    | add s _ => s
   end.

Inductive isFinite : seq -> Prop:=
|is_finite_one : forall s, isFinite (one s)
|is_finite_add : forall s sq, isFinite sq -> isFinite (add s sq).                                     


Fixpoint nth(n:nat)(sq:seq) : option State :=
  match n,sq with
  |0,  one s => Some s
  |0,  add s _  => Some s                    
  |S n', add _ tau' => nth n' tau'
  |_,_ => None
 end.           

Fixpoint suffix(n:nat)(sq:seq) : option seq :=
  match n,sq with
  |S n', add _ tau' => suffix n' tau'   
  |0, _ => Some sq
  |_,_ => None
 end.           

Lemma nth_suffix_nth  :
  forall n m s tau tau', suffix n tau = Some tau' -> nth m tau' = Some s ->
  nth (n + m) tau = Some s.                                             
Proof.
 induction n ; intros.
* cbn in *. injection H ; intros ; subst ; assumption.
* destruct tau ; cbn in * ; try discriminate.
   eapply IHn ; eauto.  
Qed.

Lemma nth_Some_suffix : forall n sq s',  nth n sq = Some s' ->  exists sq', suffix n sq = Some sq'.
Proof.
induction n  ; intros.
* destruct sq ; cbn in *.
  + injection H ; intros ; subst.  
    now exists (one s').  
  + injection H ; intros ; subst.  
      now exists (add s' sq).  
 * destruct sq ; cbn in *  ; try discriminate.
    eapply IHn ; eauto.
Qed.


Lemma head_nth_suffix : forall n sq  sq' s,
    suffix n sq = Some sq' ->
    nth n sq = Some s ->
    s = head sq'.
Proof.
induction n ; intros.
 * destruct sq ; cbn in *.
    +injection H0 ; injection H ; intros  ; subst ; reflexivity.
    +injection H0 ; injection H ; intros  ; subst ; reflexivity.
*  destruct sq ; cbn in * ; try discriminate.
   eapply IHn; eauto.
Qed.

Lemma isFinite_suffix : forall n sq sq' , isFinite sq -> suffix n sq = Some sq' -> isFinite sq'.
Proof.
 induction n ; intros.
 *   destruct sq ; cbn in *.
    +  injection H0 ; intros ; subst; constructor.
    + injection H0 ; intros ; subst; constructor.
       inversion H ; auto.      
*  destruct sq ; cbn in * ; try discriminate.
   inversion H; subst.
   eapply IHn ; eauto.   
Qed.

 CoInductive isPath: seq -> Prop :=
|path_one : forall s,  final s -> isPath (one s)
|path_add : forall s sq,  trans s (head sq) -> isPath sq -> isPath (add s sq).

Lemma isPath_suffix : forall n sq  sq' ,  isPath sq -> suffix n sq = Some sq' -> isPath sq'.
Proof.
 induction n ; intros.
* destruct sq; cbn in *.
   + injection H0 ; intros ; subst.
      inversion H0 ; subst ; auto.
   + injection H0 ; intros ; subst ; auto.
 *   destruct sq; cbn in *.
     + discriminate.
     + inversion H ; subst.
       apply IHn with sq ; auto.
Qed.

Lemma suffix_n_add : forall n tau tau' s, suffix n tau = Some tau' ->
  exists s',  suffix n (add s tau) = Some (add s' tau').
Proof.
induction n ; intros; cbn in*.
* injection H ; intros; subst ;auto.
  now exists s.
*destruct tau ; try discriminate.
 apply IHn ; auto.
Qed.

Lemma suffix_plus : forall n n' tau tau' tau'',
    suffix n tau  = Some tau' ->
    suffix n' tau' = Some tau'' ->
    suffix (n+ n') tau = Some tau''.
Proof.
  induction n; intros ; cbn in *.
  * injection H ; intros ; subst; auto.
  * destruct tau ; try discriminate.
    eapply IHn ; eauto.
Qed.

CoInductive reach : seq -> SymState-> Prop :=
|reach_now_one : forall  s (r:SymState), r s-> reach  (one s) r
|reach_now_add : forall  s tau (r:SymState), r s-> reach  (add s tau) r
|reach_later : forall s tau n tau' (r:SymState),
    suffix n tau = Some tau' ->  reach tau' r ->
    reach  (add s tau) r.


Inductive _reach (rho: seq -> SymState-> Prop):  seq -> SymState-> Prop :=
|_reach_now_one : forall  s (r:SymState), r s-> _reach rho (one s) r
|_reach_now_add : forall  s tau (r:SymState), r s-> _reach rho (add s tau) r
|_reach_later : forall s tau n tau' (r:SymState),
    suffix n tau = Some tau' ->  rho tau' r ->
    _reach rho (add s tau) r.

(* tarski strong coinduction principle for reach *)
Lemma tarski_reach : forall (X : seq -> SymState -> Prop),
   (forall tau r,  X tau r -> _reach X tau r \/ reach tau r) -> 
    forall tau r,  X tau r -> reach tau r.
Proof.
cofix tarski_reach.
intros X Hcoind  tau r HX.
destruct (Hcoind _ _ HX) ; auto.
inversion H ; subst.
* constructor ; assumption.
* constructor ; assumption.  
* econstructor 3 ; eauto.
Qed.


Lemma suffix_reach : forall n tau tau' r,
  reach tau' r -> suffix n tau = Some tau' -> reach tau r.
Proof.
induction n ; intros.
* injection H0 ; intros ; subst ; auto.
* destruct tau ; try discriminate.
  constructor 3 with 0 tau ; auto.
  eapply IHn ; eauto. 
Qed.


CoInductive reach':seq -> SymState-> Prop :=
|reach_now_one': forall s (r:SymState), r s-> reach' (one s) r
|reach_now_add': forall s tau (r:SymState), r s-> reach' (add s tau) r
|reach_later' : forall s tau r,reach' tau r -> reach' (add s tau) r.


Inductive _reach'(rho: seq -> SymState-> Prop): seq -> SymState-> Prop :=
|_reach_now_one': forall s (r:SymState), r s-> _reach' rho (one s) r
|_reach_now_add': forall s tau (r:SymState), r s-> _reach' rho (add s tau) r
|_reach_later' : forall s tau r,rho tau r -> _reach' rho (add s tau) r.

(* tarski  coinduction principle for reach' *)
Lemma tarski_reach' : forall (X : seq -> SymState -> Prop),
   (forall tau r,  X tau r -> _reach' X tau r) -> 
    forall tau r,  X tau r -> reach' tau r.
Proof.
cofix tarski_reach'.
intros X Hcoind  tau r HX.
destruct (Hcoind _ _ HX).
* constructor ; assumption.
* constructor ; assumption.
 * econstructor 3 ; eauto.
Qed.


Lemma reach_reach' : forall tau r, reach tau r -> reach' tau r.
Proof.
  intros tau r Hreach.
  apply tarski_reach' with (X := reach) ; auto ; intros.
  inversion H ; subst ; clear H.
  * constructor ; auto.
  * constructor ; auto.
  * constructor 3.
    eapply suffix_reach ; eauto.
Qed.


Lemma reach'_reach : forall tau r, reach' tau r -> reach tau r.
Proof.
  intros tau r Hreach'.
  apply tarski_reach with (X := reach') ; auto ; intros.
  left.
  inversion H ; subst ; clear H.
  * constructor ; auto.
  * constructor ; auto.
  * apply _reach_later  with(n :=0) (tau' := tau1) ; auto.
Qed.



Lemma isFinite_reach_ex : forall sq r,
    isFinite sq ->reach sq r -> exists n s,  nth n sq = Some s /\ r s.
Proof.
  intros sq r Hfin Hreach.
  induction Hfin.
  * inversion Hreach ; intros ; subst.
    exists 0,  s ; split ; auto.
  * inversion Hreach ; intros ; subst; clear Hreach.
     +  exists 0,  s ; cbn ; split ; auto.
     +  assert (Hreach1 : reach sq r).
              {
                apply suffix_reach with n tau' ; auto.
              }        
        specialize  (IHHfin Hreach1).
        destruct IHHfin as [n' [s' [Hnth Hr ]]].     
               now exists (S n'),  s'.
Qed.

Lemma nth_reach :   forall n sq (r:SymState) s,
   nth n sq = Some s -> r s -> reach sq r.
Proof.
  induction n  ; intros;  apply reach'_reach.  
* destruct sq  ; cbn in *.
  + injection H ; intros ; subst.
  constructor; auto.
  + injection H ; intros ; subst.
    constructor ; auto.
*  destruct sq  ; cbn in * ; try discriminate.
     constructor 3; auto.
    apply reach_reach' ; eapply IHn ; eauto.
Qed.


(* reachability formulas*)
Inductive Rlf:Type  :=
| rlf: forall p q: SymState, Rlf.

Notation "l  =><> r"  := (rlf  l r) (at level 100).                              

Definition lhs (f : Rlf) := match f with (l  =><> _)  => l end.
Definition rhs (f : Rlf) := match f with (_  =><> r)  => r end.



(*validity*)

Definition valid(f:Rlf) : Prop :=
  forall tau,   isPath tau ->  isFinite tau -> (lhs f) (head tau) -> reach tau (rhs f).

Definition _valid rho f := forall tau, isPath tau ->isFinite tau -> (lhs f) (head tau) ->
  _reach rho tau (rhs f).

(*some properties of validity *)


Lemma _valid_trv : forall r rho,  _valid rho (r =><> r).
 unfold _valid ; cbn ; intros.
destruct tau; constructor ; auto.
Qed.


Lemma _valid_str : forall l l' r rho,
    imp l' l -> _valid rho (l =><> r) -> _valid rho ( l'=><> r).
Proof.
  unfold _valid ; cbn ; intros.
  apply H0, H ; auto.
Qed.

Lemma _valid_dsj : forall l1 l2 r rho,
    _valid rho (l1 =><> r) -> _valid rho (l2 =><>r) -> _valid rho ((dsj l1 l2) =><>  r).
Proof.
  unfold _valid  ; cbn ; intros.
  destruct H3.
  * apply H; assumption.
  * apply H0; assumption.  
Qed.


Lemma valid_finite_ex :
  forall l r,  valid ( l =><> r) -> forall tau,  l (head tau) -> isPath tau-> isFinite tau -> exists n s,  nth n tau = Some s /\ r s.
intros. 
unfold valid in *; cbn in *.
specialize (H _ H1 H2 H0).
specialize (isFinite_reach_ex  _ _ H2 H)  ; intro Hex.
 destruct Hex as [n [s [Hnth Hr]]]. 
now exists n, s.
Qed.


Lemma ex_finite_valid :
  forall (l r:SymState),
    (forall  tau,  l (head tau) -> isPath tau-> isFinite tau -> exists n s,  nth n tau = Some s /\ r s)
    ->
    valid ( l =><> r).
Proof.
  unfold valid ; intros ; cbn in *.
  specialize (H _ H2 H0 H1).
  destruct H as [n [s [Hnth Hr]]].    
    eapply nth_reach ; eauto.
Qed.


Lemma valid_iff_finite_ex :  forall (l r:SymState),
       valid ( l =><> r) <->   (forall  tau,  l (head tau) -> isPath tau-> isFinite tau -> exists n s,  nth n tau = Some s /\ r s).
Proof.
split.
 * apply valid_finite_ex.  
 * apply ex_finite_valid.
Qed.


Lemma valid_tra : forall l r m,  valid(l =><> m) -> valid (m =><> r) -> valid(l =><> r).
Proof.
  intros.
  rewrite valid_iff_finite_ex in * ; intros.
 specialize (H _ H1 H2 H3).
 destruct H as [n [s' [Hnth Hm]]].
 specialize (nth_Some_suffix _ _ _ Hnth) ; intro Hexsuf.
 destruct Hexsuf as [tau' Hsuf]. 
 specialize (head_nth_suffix _ _ _ _ Hsuf Hnth) ; intro Hhds'  ; subst.
specialize (isFinite_suffix _ _ _ H3 Hsuf) ; intro Hfinsuf.
specialize (isPath_suffix  _ _ _ H2 Hsuf) ; intro Hpathsuf.
specialize (H0 _ Hm Hpathsuf Hfinsuf).
 destruct H0 as [n' [s [Hnth' Hr]]].
exists (n + n'), s ; split ; auto.
eapply nth_suffix_nth ; eauto. 
Qed.



    
Inductive _proof (H rho: Rlf -> Prop)  : Rlf -> Prop :=
|Hyp : forall f,  H f -> _proof H rho f  
|Trv : forall r,  _proof H rho ( r =><> r)
|Str : forall l l' r,
    imp l l' ->
    _proof H rho (l' =><> r)  ->
    _proof H rho  (l =><> r)               
|Spl : forall l1 l2 r,
    _proof H rho (l1 =><> r) ->
    _proof H rho (l2 =><> r) ->
    _proof H rho ((dsj l1 l2) =><> r)
|Tra :
    forall l m r,
    valid (l =><> m) ->
    _proof H rho (m =><> r) ->
    _proof H rho (l =><> r)
|Stp : forall l r , rho (symTrans l =><> r) ->
    imp (cnj l final) bot ->
    _proof H rho (l =><> r).      
 
(*
CoInductive proof(H: Rlf -> Prop) : Rlf -> Prop :=
|prefix : forall f,  _proof H (proof H) f -> proof H f.
 *)

Lemma _proof_mono : forall H,  monotone1 (_proof H).
Proof.
  unfold monotone1; intros.
  induction IN.
  * constructor 1 ; auto.
  * constructor 2.
  * econstructor 3  ; eauto.
  * constructor 4 ; auto.
  * econstructor 5 ; eauto.
  * constructor 6 ; auto.
Qed.

Definition proof H := paco1 (_proof H) bot1.


Lemma soundness_aux :
  forall (H : Rlf -> Prop),
    (forall f : Rlf, H f -> valid f) ->
       forall l r, proof H (l =><> r) ->   
           forall tau, 
               l (head tau) -> isPath tau -> isFinite tau
  ->
  _reach
    (fun (tau' : seq) (r' : SymState) =>
     exists l' : SymState,
       proof H (l' =><> r') /\
       l' (head tau') /\ isPath tau' /\ isFinite tau') tau r \/ reach tau r.
Proof.
intros H Hval l r Hproof.
punfold Hproof ; subst ;  intros; [| apply _proof_mono].
dependent induction Hproof.
  * right.
    apply (Hval _ H0 _ H2 H3 H1). 
  * right.
    inversion H2 ; subst.
    + constructor ; assumption.    
    + constructor; assumption.
    * 
        eapply IHHproof ; eauto.
  * destruct H0.
   + eapply IHHproof1 ; eauto.
   + eapply IHHproof2 ; eauto.
  * specialize (H0 _ H2 H3 H1) ; cbn in *.
  inversion H0 ; subst ; cbn in *.
   + eapply IHHproof ; eauto.
   + eapply IHHproof ; eauto.
   +
     rename tau0 into tau.
     inversion H3 ; subst.
     generalize (isFinite_suffix _ _ _  H7 H4) ; intro Hfintau'.
     destruct (isFinite_reach_ex _ _  Hfintau' H5) as (n' & s' & Hnth & Hm).
     specialize (nth_Some_suffix _ _ _ Hnth) as (tau'' & Hsuf).
     specialize (suffix_plus _ _ _ _ _  H4 Hsuf) ; intro Hsuf''.
     specialize (head_nth_suffix _ _  _ _ Hsuf Hnth) ; intro ; subst.
     inversion H2 ; subst.
     specialize (isPath_suffix _ _ _  H10 Hsuf'') ; intro HisPath''.
     specialize (isFinite_suffix _ _ _  H7 Hsuf'') ; intro HisFinite''.
     destruct (IHHproof Hval  _ _ (eq_refl)_ Hm HisPath'' HisFinite'').
     - left.
       econstructor 3 ; eauto.
       exists m.  split.
       **
         pfold ; auto.
       **  repeat split ; auto.
     - right.
       econstructor 3 ; eauto.
 * inversion H3 ; subst  ; cbn in *.
   + exfalso.
     apply (H1 s) ; split ; auto.
   + left.
     assert (suffix 0 sq = Some sq) ; auto.
     econstructor 3 ; eauto.
     exists (symTrans l); split ; auto.
     - inversion H0.
       ** assumption.
       ** inversion H8.
     - repeat split  ; auto.
       ** exists s; split; auto.
       ** inversion H4 ; auto.       
Qed.

Theorem soundness : forall (H: Rlf -> Prop) f,
    (forall f, H f -> valid f) ->
                                proof H f -> valid f.
Proof.
  intros H f Hval Hproof tau HisPath Hfin Hinit.
  apply tarski_reach with
      (X := fun tau' r' =>
              exists l', proof H (l' =><> r') /\
                         l' (head tau') /\ isPath tau' /\isFinite tau').
  
  *  clear - Hval.
     intros .
     destruct H0 as (l' & Hproof & Hinit & HisPath & Hfin).
     eapply soundness_aux ; eauto.
    
  * destruct f as (l & r) ; cbn in *.
    now exists l.
Qed.

Lemma proof_tarski : forall (H X : Rlf -> Prop),
  (forall f,  X f -> _proof H X f) -> 
  (forall f,  X f -> proof H f).
Proof.
intros H X.  
pcofix proof_tarski.
intros.
pfold.
 eapply _proof_mono ; auto.
Qed.

  Section CompleteStrategy.
(* let's assume three symbolic states l, q, r *)  
  Variable l  r  q : SymState.
(* if the folllowing hypotheses hold on l, q and r, then (l =><> r) has a proof and thus is valid *)
  Hypothesis Hinit : imp  l (dsj q r).
  Hypothesis Hnonblock : imp (cnj q final) bot.
  Hypothesis HquasiInv : imp (symTrans q) (dsj q r).


Lemma strategy: proof bot1 (l =><> r).
Proof.
set (FF :=  fun f =>
                  (f = (l =><> r))\/
                  (f = ((dsj q r) =><> r)) \/
                  (f = (symTrans q =><> r)) \/
                  (f = (q =><> r)) \/
                  (f = (r =><> r))).
 apply proof_tarski with (X:=  FF).
 intros  f  [Hf | [  Hf| [ Hf |[ Hf | Hf]]]] ; subst.
 *apply Str with (dsj q r) ; auto.
  apply Spl; [ | apply Trv].
  apply Stp ; auto ; unfold FF ;intuition.
 *apply Spl; [ | apply Trv].
  apply Stp ; auto ; unfold FF ;intuition.
 * apply Str with (dsj q r) ; auto.
  apply Spl; [ | apply Trv].
  apply Stp ; auto ; unfold FF ;intuition.
 * apply Stp ; auto ; unfold FF ;intuition.
 * apply Trv.
 * unfold FF ; intuition.    
Qed.

  End CompleteStrategy.


(* we now prove that a "q" like in CompleteStrategy always exists for valid rlfs *)
Definition coReach(r:SymState) : SymState :=
  fun s => ~ r s /\ forall tau, s = head tau-> isPath tau -> isFinite tau -> reach tau r.


Lemma init : forall f, valid f -> imp (lhs f) (dsj (coReach (rhs f)) (rhs f)).
Proof.
  unfold valid, coReach, imp, dsj; intros.
  destruct f as (l,r) ; cbn in *.
  destruct (classic (r s)) ; auto.
  left ; split; auto ; intros ; subst.
   apply H; auto.    
Qed.




Lemma nonBlock : forall r,  imp (cnj (coReach r) final) bot.
Proof.
  unfold imp, cnj, coReach, bot ; intros.
  do 2 destruct H.
  assert (isPath (one s)).
  * constructor ; auto.
  *  specialize (H1 (one s) (refl_equal _) H2).
     assert (Hfin : isFinite (one s)). {constructor. } 
     specialize (H1 Hfin).          
     inversion H1; subst ; intuition.
Qed.





Lemma quasiInv : forall r, imp (symTrans (coReach r)) (dsj  (coReach r) r).
Proof.
  unfold imp, symTrans, coReach, dsj; intros.
  destruct H as [s' [[Hnr Hreach] Htrans]].
  destruct (classic (r s)) ; auto.
  left  ; split ; auto ; intros ; subst.
  assert (reach (add s' tau) r).
  * apply Hreach ; try reflexivity ; constructor ; auto.
  * apply reach'_reach.
    apply reach_reach' in H0.
    inversion H0 ; subst ; intuition.
Qed.

Theorem completeness : forall f,  valid f -> proof bot1 f.
  intros.
  destruct f as (l,r).
  apply strategy with (coReach r).
  * apply (init (l =><> r))  ; auto.
  *  apply nonBlock.
  *  apply quasiInv.
Qed.
