Require Import List.
Require Import Classical.
Import ListNotations.

Parameter State : Type.
Parameter trans : State -> State -> Prop.

Definition SymState := State -> Prop.
Definition top : SymState := fun _ => True.
Definition bot : SymState := fun _ => False.
Definition cnj(p q :SymState) := fun s => p s /\ q s.
Definition dsj(p q :SymState) := fun s => p s \/ q s.
Definition neg (p:SymState) := fun s => ~ p s.
Definition imp(p q :SymState) := forall s, p s -> q s.

(* implication is transitive *)
Lemma imp_trans : forall p q r, imp p q -> imp q r -> imp p r.
Proof.
  unfold imp  ; intros.
  apply H0, H ; assumption.
Qed.  
  (* symbolic transition function *)
Definition symTrans(q:SymState) := fun s => exists s', q s' /\ trans s' s.

(* final states *)
Definition final : SymState := fun s => forall s', ~trans s s'.

(* sequences and paths *)
CoInductive seq  : Type :=
| one : forall s:State , seq
| add : State -> seq  -> seq.                                          

Definition head(sq : seq) : State :=
  match sq with
    | one s => s
    | add s _ => s
   end.

 CoInductive isPath: seq -> Prop :=
|path_one : forall s,  final s -> isPath (one s)
|path_add : forall s sq,  trans s (head sq) -> isPath sq -> isPath (add s sq).

(* when does a sequence "reach" a symbolic state *)
CoInductive reach : seq -> SymState -> Prop :=
|reach_now_one : forall s (r:SymState),  r s-> reach (one s) r
|reach_now_add : forall s tau (r:SymState),   r s-> reach (add s tau) r
|reach_later : forall s tau (r:SymState),  reach tau r -> reach (add s tau) r.

(* functional corresponding to coinductive predicate reach *)
Inductive _reach(X:seq -> SymState -> Prop) :   seq -> SymState -> Prop :=
|_reach_now_one : forall s (r:SymState),  r s-> _reach X (one s) r
|_reach_now_add : forall s tau (r:SymState),   r s-> _reach X (add s tau) r
|_reach_later : forall s tau (r:SymState),  X tau r -> _reach X (add s tau) r.

(* tarski coinduction principle for reach *)
Lemma tarski_reach : forall (X : seq -> SymState -> Prop),
   (forall tau r,  X tau r -> _reach X tau r) -> 
    forall tau r,  X tau r -> reach tau r.
Proof.
cofix tarski_reach.
intros X Hcoind  tau r HX.
destruct (Hcoind _ _ HX).
* constructor ; assumption.
* constructor ; assumption.  
* constructor 3; eapply tarski_reach ; eauto.
Qed.

  (* reachability-logic forrmulas*)
Inductive Rlf:Type  :=
| rlf: forall p q: SymState, Rlf.

Notation "l  =><> r"  := (rlf  l r) (at level 100).                              

Definition lhs (f : Rlf) := match f with (l  =><> _)  => l end.
Definition rhs (f : Rlf) := match f with (_  =><> r)  => r end.



(*validity*)
Definition valid(f:Rlf) : Prop :=
  forall tau,   isPath tau ->  (lhs f) (head tau) -> reach tau (rhs f).


(* a one-rule proof system  *)
CoInductive proof:  Rlf  -> Prop :=
|Stp : forall (l l'  r: SymState),
    imp l (dsj l' r) -> (* i.e., l' over-approximates l \ r*)
    imp (cnj l' final) bot ->
    proof  (symTrans l' =><> r) ->
    proof (l =><> r).

(* soundness: all proved formulas are valid*)
Theorem soundness : forall f, proof f -> valid f.
Proof.
   intros (l & r) Hproof tau HisPath Hinit; cbn in *.
   assert (exists l : SymState, proof (l =><> r) /\ isPath tau /\ l (head tau)); [firstorder |].
   apply tarski_reach with
       (X := fun tau r =>  exists l, proof (l =><> r) /\ isPath tau /\ l (head tau)) ; auto.
   clear.   
   intros tau r (l' & Hproof & HisPath & Hl').
   inversion Hproof; subst; clear Hproof.
   inversion HisPath; subst; clear HisPath.
   destruct (H1 _ Hl'); auto ; cbn in *.
   * exfalso; apply (H2 s) ; split; auto.
   * constructor ; auto.
   * destruct (H1 _ Hl').
     +  constructor 3.
        exists (symTrans l'0) ; repeat split  ; auto.
        now exists s.
     + constructor 2 ; auto.
Qed.


(* functional corresponding to proof *)
Inductive _proof(X: Rlf -> Prop) :  Rlf  -> Prop :=
|_Stp : forall (l l'  r: SymState),
    imp l (dsj l' r) -> 
    imp (cnj l' final) bot ->
    X (symTrans l' =><> r) ->
    _proof X (l =><> r).

(* tarski coinduction principle for proof *)
Lemma tarski_proof : forall (X : Rlf -> Prop),
   (forall f,  X f -> _proof X f) -> 
    forall f,  X f -> proof f.
Proof.
cofix tarski_proof.
intros X Hcoind  f HX.
destruct (Hcoind _ HX).
econstructor  ; eauto.
Qed.


Section ProvingSomeRLFormulas.
  Variable f : Rlf.
  Variable q : SymState.
  
(* if the folllowing hypotheses hold on l, q and r, then (l =><> r) has a proof and thus is valid *)
  Hypothesis Hinit : imp  (lhs f) (dsj q (rhs f)).
  Hypothesis Hnonblock : imp (cnj q final) bot.
  Hypothesis HquasiInv : imp (symTrans q) (dsj q (rhs f)).

Lemma strategy : proof f.
Proof.
  assert(Hex : exists l', imp  (lhs f) l' /\ imp l' (dsj q (rhs f)) /\ imp (cnj q final) bot /\ imp (symTrans q) (dsj q (rhs f))) ;
    [firstorder|].
  apply tarski_proof with
      (X := fun f => exists l',
          imp (lhs f) l' /\
          imp l' (dsj q (rhs f)) /\
          imp (cnj q final) bot /\
          imp (symTrans q) (dsj q (rhs f))) ; auto.
clear. 
 intros (l & r) Hmatch ; cbn in *.
 destruct Hmatch as (l' & Himp' & Hinit' & Hnonblock' & HquiasiInv').
 apply _Stp with q ; auto.
 * apply imp_trans with (q := l') ; auto.
 * cbn.
   now exists (dsj q r).
 Qed.


  
End ProvingSomeRLFormulas.


(* we now prove that a "q" llike above always exists for valid rechability formulas *)
Definition coReach(r:SymState) : SymState :=
  fun s => ~ r s /\ forall tau, s = head tau-> isPath tau -> reach tau r.

Lemma init : forall f, valid f -> imp (lhs f) (dsj (coReach (rhs f)) (rhs f)).
Proof.
  unfold valid, coReach, imp, dsj; intros.
  destruct f as (l,r) ; cbn in *.
  destruct (classic (r s)) ; auto.
  left ; split; auto ; intros ; subst.
   apply H; auto.    
Defined.

Lemma nonBlock : forall r,  imp (cnj (coReach r) final) bot.
Proof.
  unfold imp, cnj, coReach, bot ; intros.
  do 2 destruct H.
  assert (isPath (one s)).
  * constructor ; auto.
  *  specialize (H1 (one s) (refl_equal _) H2).
     inversion H1; subst ; intuition.
Defined.

Lemma quasiInv : forall r, imp (symTrans (coReach r)) (dsj  (coReach r) r).
Proof.
  unfold imp, symTrans, coReach, dsj; intros.
  destruct H as [s' [[Hnr Hreach] Htrans]].
  destruct (classic (r s)) ; auto.
  left  ; split ; auto ; intros ; subst.
  assert (reach (add s' tau) r).
  * apply Hreach ; try reflexivity.
    constructor ; auto.
  *  inversion H0 ; subst ; intuition.
Defined.

Theorem completeness : forall f,  valid f -> proof f.
  intros.
  destruct f as (l,r).
  apply strategy with (coReach r).
  * apply (init (l =><> r))  ; auto.
  *  apply nonBlock.
  *  apply quasiInv.
 Defined.    
