Require Import Omega ArithRing.
Require Export SigTS SigComp.
Require Import StructComp. (* also imports ReachabilityLogic *)

(* component's state and transition relation *)
Definition State' := (nat*nat*nat)%type.

Inductive trans' : State' -> State' -> Prop:=
  |loop' : forall m i s, i < m -> trans'(m,i,s)(m, i+1, s+i+1).

Definition final' : State' -> Prop :=  
   fun s => forall s', ~ trans' s s'.



Module CStatesTransitionsMod <: SigTSMod.
  Definition State := State'.
  Definition trans := trans'.
  Definition final := final'.
  Lemma final_is_final : forall s, final s <-> forall s', ~ trans s s'.
  Proof.
    split ; intros ; auto.
Qed.
           
End CStatesTransitionsMod.

(* main state and transition relation *)
Inductive Loc := l0 | l1 | l2.

Definition State := (Loc*nat*nat*nat)%type.

Inductive trans : State -> State -> Prop:=
  |init : forall m i s, trans (l0,m,i,s) (l1,m,0,0)
  |loop : forall m i s m' i' s',  trans'(m,i,s)(m', i', s') -> trans(l1, m,i,s)(l1,m',i',s')
  |stop : forall m i s,  i >= m -> trans(l1,m,i,s)(l2,m,i,s).                         

Definition final : State -> Prop :=  
   fun s => forall s', ~ trans s s'.


Module StatesTransitionsMod <: SigTSMod. 
  Definition State := State.    
  Definition trans := trans.
  Definition final := final.
  Lemma final_is_final : forall s, final s <-> forall s', ~ trans s s'.
  Proof.
    split ; intros ; auto.
Qed.
End StatesTransitionsMod.



(* showing component relation*)
Module CompMod <: SigCompMod StatesTransitionsMod CStatesTransitionsMod.
  Definition State' := CStatesTransitionsMod.State.
  Definition trans' := CStatesTransitionsMod.trans.

  Definition State := StatesTransitionsMod.State.
  Definition trans := StatesTransitionsMod.trans.
  
  Definition state_map : State' -> State := 
  fun (s': State') => match s' with (m,i,s) => (l1,m,i,s) end.

  Lemma state_map_inj : forall s'1 s'2 ,
      state_map s'1 = state_map s'2 -> s'1 = s'2.
  Proof.
    intros  ((ma,ia),sa) (( mb,ib),sb) Heq; cbn in *.
    inversion Heq ; subst ; auto.
  Qed.

  Lemma subtrans :
    forall s'1 s'2, trans' s'1 s'2 -> trans (state_map s'1) (state_map s'2). 
  Proof.
    intros  ((ma,ia),sa) (( mb,ib),sb) Htrans'.
     
   unfold state_map ; cbn.
   inversion Htrans' ; subst.
   constructor 2 ; auto.
  Qed.

 Lemma full_subtrans : forall s1 s2 s'1 s'2,
   trans s1 s2 -> state_map s'1 = s1 ->   state_map s'2 = s2 ->  trans' (s'1) (s'2).
 Proof.
   intros (((la,ma),ia),sa) (((lb,mb),ib),sb) ((m,i),s) ((m',i'),s') Htrans Hsm1 Hsm2.
   cbn in Hsm1, Hsm2 ; inversion Hsm1; inversion Hsm2 ; subst.
   inversion Htrans ; subst.
   inversion H0 ; subst.
   constructor ; auto.
 Qed.

 Lemma clean_border :
forall s'1 s2, 
    trans (state_map s'1)  s2 ->
    ~isImage state_map s2 -> final' s'1.
 Proof.
   intros ((ma,ia),sa) (((lb,mb),ib),sb) Htrans Hnex ((m',i'),s').
   cbn in Htrans; inversion Htrans ; subst.
   
   * exfalso.
     apply Hnex.
     now exists (mb, ib, sb).
 * intros Habs.
    inversion Habs ; subst ; omega.                       
Qed.
End CompMod.

Module InstCompMod :=
  StructCompMod StatesTransitionsMod CStatesTransitionsMod CompMod.

Import InstCompMod.

(* showing validity of a formula  l' =><> r' in the component *)
Definition l' : RL'.SymState := 
   fun s => match s with (_,i,s) =>
                         i = 0 /\ s = 0 end.

Definition r' : RL'.SymState := 
   fun s => match s with (m,i,s) =>
       i = m  /\ 2* s = m * (m+1) end.

Definition q': RL'.SymState := 
fun s => match s with (m,i,s) =>
                      i < m /\ 2*s = i * (i+1)  end.

Lemma Init : RL'.imp l' (RL'.dsj q' r'). 
Proof.
unfold RL'.imp, RL'.dsj, q', l', r'.
intros ((m,i),s) (H1 & H2) ; subst.
destruct m.
* right ; split ; auto.
* left ; split ; omega.
Qed.

Lemma QuasiInv : RL'.imp (RL'.symTrans q')  (RL'.dsj q' r').
Proof.
unfold RL'.imp, RL'.symTrans, RL'.dsj, q', l', r'.
intros ((m,i),s) H.
destruct H as (((m',i'),s') & (Hlt & Heq)  & Htrans).
inversion Htrans ; subst ; clear Htrans.
destruct (le_lt_dec m (i' + 1)).
* right ; split ; auto ; try omega.
  assert (i' +1  = m) ; try omega; subst.
  replace (2 * (s' + i' + 1) ) with (2*s' + 2* i' + 2) ; subst ; try omega.
  rewrite Heq.
  ring.
*  left ; split; auto.
    replace (2 * (s' + i' + 1) ) with (2*s' + 2* i' + 2) ; subst ; try omega.
  rewrite Heq.
  ring.
Qed.

Lemma NonBlock : RL'.imp (RL'.cnj q' final') RL'.bot.
Proof.
  unfold RL'.imp, RL'.cnj, final', RL'.bot, q', l', r'.
  intros ((m,i),s) ((H1 & H2) & Htrans).
  apply (Htrans (m,i+1, s+i+1)); clear Htrans.
  constructor; auto.
Qed.

  Lemma isValid' : RL'.valid (RL'.rlf l' r').
Proof.
apply RL'.soundness.
apply (RL'.strategy) with q'.
* apply Init.
* apply NonBlock.
* apply QuasiInv.
Qed.

(* now showing validity of corresponding formula l =><> r on whole system*)

Definition l : RL.SymState := 
   fun s => match s with (l,_,i,s) =>
                         l = l1 /\i = 0 /\ s = 0 end.

Definition r : RL.SymState := 
   fun s => match s with (l,m,i,s) =>
                         l = l1 /\  i = m /\ 2* s = m * (i+1) end.

(* correspondence between l and l'*)
Lemma l_symState_map_l' : forall s, l s -> (symState_map l') s.
Proof.
intros (((la,ma),sa),ia) Hls.  
unfold l in Hls ; cbn in Hls.
destruct Hls as (H1 & H2 & H3) ; subst.
exists (ma,0,0) ; split ; auto.
repeat split; auto.
Qed.


(* correspondence between r' and r*)
Lemma r'_symState_map_r : forall s,  (symState_map r') s -> r s.
Proof.
  intros (((la,ma),sa),ia) Hrs.
  destruct Hrs as (((m',s'),i') & H1 & H2) ; subst.
  unfold CompMod.state_map in H2 ; injection H2 ; intros ; subst.
  unfold l' in H1 ; cbn in H1.
  destruct H1; subst.
  unfold r; repeat split; auto.
Qed.


Lemma isAlsoValid : RL.valid (RL.rlf l r).
Proof.
generalize (ValidComp _ isValid') ; intro Hval.
cbn in Hval.
rewrite RL.valid_iff_finite_ex in *; intros.
apply l_symState_map_l' in H.
destruct (Hval _ H H0 H1)  as (n & s & Hnth & Hsym).
exists n, s ; repeat split ; auto.
apply r'_symState_map_r ; assumption.
Qed.


